{
(*
  Copyright 2002-2004 Sebastien Ailleret
  Copyright 2004 Martin Jambon

  This file is distributed under the terms of the GNU Public License
  http://www.gnu.org/licenses/gpl.txt
*)

open Lexing

let keywords = [
  "and";
  "as";
  "asr";
  "assert";
  "begin";
  "class";
  "constraint";
  "do";
  "done";
  "downto";
  "else";
  "end";
  "exception";
  "external";
  "false";
  "for";
  "fun";
  "function";
  "functor";
  "if";
  "in";
  "include";
  "inherit";
  "initializer";
  "land";
  "lazy";
  "let";
  "lor";
  "lsl";
  "lsr";
  "lxor";
  "match";
  "method";
  "mod";
  "module";
  "mutable";
  "new";
  "object";
  "of";
  "open";
  "or";
  "private";
  "rec";
  "sig";
  "struct";
  "then";
  "to";
  "true";
  "try";
  "type";
  "val";
  "virtual";
  "when";
  "while";
  "with" ]

let depth = ref 0

let is_keyword =
  let tbl = Hashtbl.create 100 in
  List.iter (fun key -> Hashtbl.add tbl key ()) keywords;
  Hashtbl.mem tbl

let tokenify s =
  if is_keyword s then `Keyword s
  else `Token s

let buf = Buffer.create 256

let token_list = ref []
let add_token x = token_list := x :: !token_list

}

let upper = ['A'-'Z' '\192'-'\214' '\216'-'\222']
let lower = ['a'-'z' '\223'-'\246' '\248'-'\255']
let digit = ['0'-'9']
let identchar = upper | lower | digit | ['_' '\'']
let hex = ['0'-'9' 'a'-'f' 'A'-'F']
let oct = ['0'-'7']
let bin = ['0'-'1']

let operator_char = 
  [ '!' '$' '%' '&' '*' '+' '-' '.' '/' ':' '<' '=' '>' '?' '@' '^' '|' '~']
let infix_symbol =
  ['=' '<' '>' '@' '^' '|' '&' '+' '-' '*' '/' '$' '%'] operator_char*
let prefix_symbol = ['!' '?' '~'] operator_char*

rule token = parse
| "(*"
    { Buffer.clear buf;
      Buffer.add_string buf "(*";
      depth := 1;
      comment lexbuf;
      add_token (`Comment (Buffer.contents buf));
      token lexbuf }
| '"'
    { Buffer.clear buf;
      Buffer.add_char buf '"';
      string lexbuf;
      add_token (`String (Buffer.contents buf));
      token lexbuf }
| (('`' (upper | lower)) | upper) identchar *
      { add_token (`Construct (lexeme lexbuf));
	token lexbuf }
| lower identchar *
      { add_token (tokenify (lexeme lexbuf));
	token lexbuf  }

| "!=" | "#" | "&" | "&&" | "(" | ")" | "*" | "+" | "," | "-"
| "-." | "->" | "." | ".. :" | "::" | ":=" | ":>" | ";" | ";;" | "<"
| "<-" | "=" | ">" | ">]" | ">}" | "?" | "??" | "[" | "[<" | "[>" | "[|"
| "]" | "_" | "`" | "{" | "{<" | "|" | "|]" | "}" | "~"

    { add_token (`Keyword (lexeme lexbuf));
      token lexbuf }

| prefix_symbol | infix_symbol
      { add_token (`Token (lexeme lexbuf));
	token lexbuf }

| "'\n'" | "'\r\n'" 
      { List.iter add_token [`String "'"; `Newline; `String "'"];
	token lexbuf }
| "'\\\n'" | "'\\\r\n'"
      { List.iter add_token [`String "'\\"; `Newline; `String "'"];
	token lexbuf }
| "'" ([^'\'''\\'] | '\\' (_ | digit digit digit | 'x' hex hex)) "'"
      { add_token (`String (lexeme lexbuf));
	token lexbuf }

| '\r'? '\n'
    { add_token `Newline;
      token lexbuf }
| '\t'
    { add_token `Tab;
      token lexbuf }
| eof
    { let l = List.rev !token_list in
      token_list := [];
      l }
| ' '+
    { add_token (`Token (lexeme lexbuf));
      token lexbuf }

| '-'? (digit (digit | '_')*
       | ("0x"| "0X") hex (hex | '_')*
       | ("0o"| "0O") oct (oct | '_')*	
       | ("0b"| "0B") bin (bin | '_')* )

| '-'? digit (digit | '_')* ('.' (digit | '_')* )? 
      (['e' 'E'] ['+' '-']? digit (digit | '_')* )? 
| _
    { add_token (`Token (lexeme lexbuf));
      token lexbuf }

and comment = parse
| "(*"
    { incr depth;
      Buffer.add_string buf "(*";
      comment lexbuf }
| "*)"
    { decr depth; 
      Buffer.add_string buf "*)";
      if (!depth > 0) then comment lexbuf }
| '"'
    { Buffer.add_char buf '"';
      string lexbuf;
      comment lexbuf }
| eof
    { () }
| '\r'? '\n'
    { add_token (`Comment (Buffer.contents buf));
      add_token `Newline;
      Buffer.clear buf;
      comment lexbuf }
| '\t'
    { add_token (`Comment (Buffer.contents buf));
      add_token `Tab;
      Buffer.clear buf;
      comment lexbuf }
| _
    { Buffer.add_string buf (lexeme lexbuf);
      comment lexbuf }

and string = parse
| '"'
    { Buffer.add_char buf '"' }
| "\\\\"
| '\\' '"'
    { Buffer.add_string buf (lexeme lexbuf);
      string lexbuf }
| eof
    { () }
| '\r'? '\n'
    { add_token (`String (Buffer.contents buf));
      add_token `Newline;
      Buffer.clear buf;
      string lexbuf }
| '\t'
    { add_token (`String (Buffer.contents buf));
      add_token `Tab;
      Buffer.clear buf;
      string lexbuf }
| _
    { Buffer.add_string buf (lexeme lexbuf);
      string lexbuf }

{
  type token = [ `Comment of string
	       | `Construct of string
	       | `Keyword of string
	       | `Newline
	       | `String of string
	       | `Tab
	       | `Token of string ]
  
  let parse = token
  let string s = let lexbuf = Lexing.from_string s in token lexbuf
  let channel ic = let lexbuf = Lexing.from_channel ic in token lexbuf
  let file s = 
    let ic = open_in s in
    let l = channel ic in
    close_in ic;
    l
}
